searchNodes=[{"doc":"This module contains top-level functions useful to validate, manipulate and inspect Spike forms. To learn about how to create Spike forms, see Spike.Form .","ref":"Spike.html","title":"Spike","type":"module"},{"doc":"Updates embeds_many association on the top-level Spike form, appending the new item at the end of the existing list. iex&gt; form = Test.ComplexForm . new ( %{ company : %{ name : &quot;ACME Corp.&quot; } , partners : [ ] } ) iex&gt; form . partners [ ] iex&gt; form = Spike . append ( form , :partners , %{ name : &quot;Hubert&quot; } ) iex&gt; form . partners |&gt; Enum . map ( &amp; &amp;1 . name ) [ &quot;Hubert&quot; ] iex&gt; form = Spike . append ( form , :partners , %{ name : &quot;Wojciech&quot; } ) iex&gt; form . partners |&gt; Enum . map ( &amp; &amp;1 . name ) [ &quot;Hubert&quot; , &quot;Wojciech&quot; ]","ref":"Spike.html#append/3","title":"Spike.append/3","type":"function"},{"doc":"Updates embeds_many association on given Spike form, appending the new item at the end of the existing list. iex&gt; form = Test.ComplexForm . new ( %{ company : %{ name : &quot;ACME Corp.&quot; } , partners : [ ] } ) iex&gt; form . partners [ ] iex&gt; form = Spike . append ( form , form . ref , :partners , %{ name : &quot;Hubert&quot; } ) iex&gt; form . partners |&gt; Enum . map ( &amp; &amp;1 . name ) [ &quot;Hubert&quot; ] iex&gt; form = Spike . append ( form , form . ref , :partners , %{ name : &quot;Wojciech&quot; } ) iex&gt; form . partners |&gt; Enum . map ( &amp; &amp;1 . name ) [ &quot;Hubert&quot; , &quot;Wojciech&quot; ]","ref":"Spike.html#append/4","title":"Spike.append/4","type":"function"},{"doc":"Deletes a form from a Spike form by it's ref . Useful to remove items from embeds_many or embeds_one . iex&gt; form = Test.ComplexForm . new ( %{ company : %{ name : &quot;ACME Corp.&quot; } , partners : [ %{ name : &quot;John&quot; } ] } ) iex&gt; form = Spike . delete ( form , form . company . ref ) iex&gt; form = Spike . delete ( form , form . partners |&gt; hd ( ) |&gt; Map . get ( :ref ) ) iex&gt; form . company nil iex&gt; form . partners [ ]","ref":"Spike.html#delete/2","title":"Spike.delete/2","type":"function"},{"doc":"Given a Spike form, returns a map of fields that have been updated since the form form was created, or since the last time Spike.make_pristine/1 was called. iex&gt; form = Test.SimpleForm . new ( %{ first_name : &quot;Spike&quot; , last_name : &quot;Spiegel&quot; } ) iex&gt; Spike . dirty_fields ( form ) %{ } iex&gt; form = Spike . update ( form , form . ref , %{ first_name : &quot;Jet&quot; , last_name : &quot;Black&quot; } ) iex&gt; Spike . dirty_fields ( form ) %{ form . ref =&gt; [ :first_name , :last_name ] } iex&gt; form = Spike . make_pristine ( form ) iex&gt; Spike . dirty_fields ( form ) %{ }","ref":"Spike.html#dirty_fields/1","title":"Spike.dirty_fields/1","type":"function"},{"doc":"Given a Spike form, returns a map of errors, where fields are refs of all forms that build the form, and each field is a list of errors. Returns empty map in case the form is valid. iex&gt; form = Test.SimpleForm . new ( %{ } ) iex&gt; Spike . errors ( form ) [ form . ref ] %{ accepts_conditions : [ acceptance : &quot;must be accepted&quot; ] , first_name : [ presence : &quot;must be present&quot; ] } iex&gt; form = Spike . update ( form , form . ref , %{ first_name : &quot;Spike&quot; , last_name : &quot;Spiegel&quot; , accepts_conditions : &quot;1&quot; } ) iex&gt; Spike . errors ( form ) %{ } iex&gt; form = Test.ComplexForm . new ( %{ company : %{ } , partners : [ ] } ) iex&gt; Spike . errors ( form ) %{ form . company . ref =&gt; %{ name : [ presence : &quot;must be present&quot; ] } , form . ref =&gt; %{ accepts_conditions : [ acceptance : &quot;must be accepted&quot; ] } }","ref":"Spike.html#errors/1","title":"Spike.errors/1","type":"function"},{"doc":"Returns true if given form, identified by ref has error on given field . iex&gt; form = Test.SimpleForm . new ( %{ first_name : &quot;Jet&quot; , accepts_conditions : &quot;0&quot; } ) iex&gt; Spike . has_errors? ( form , form . ref , :first_name ) false iex&gt; Spike . has_errors? ( form , form . ref , :accepts_conditions ) true","ref":"Spike.html#has_errors?/3","title":"Spike.has_errors?/3","type":"function"},{"doc":"Returns true if given form, identified by ref has specified error message on given field . iex&gt; form = Test.SimpleForm . new ( %{ first_name : &quot;Jet&quot; , accepts_conditions : &quot;0&quot; } ) iex&gt; Spike . has_errors? ( form , form . ref , :accepts_conditions , &quot;must be accepted&quot; ) true iex&gt; Spike . has_errors? ( form , form . ref , :accepts_conditions , &quot;is invalid&quot; ) false","ref":"Spike.html#has_errors?/4","title":"Spike.has_errors?/4","type":"function"},{"doc":"Given a Spike form, returns a map of errors, where fields are Strings representing where in nested set of forms errors happened and values are errors. Returns empty map in case the form is valid. iex&gt; form = Test.ComplexForm . new ( %{ company : %{ } , partners : [ %{ } ] } ) iex&gt; Spike . human_readable_errors ( form ) %{ &quot;accepts_conditions&quot; =&gt; [ &quot;must be accepted&quot; ] , &quot;company.name&quot; =&gt; [ &quot;must be present&quot; ] }","ref":"Spike.html#human_readable_errors/1","title":"Spike.human_readable_errors/1","type":"function"},{"doc":"Given Spike form, marks all of it's fields as dirty, recursively. iex&gt; form = Test.ComplexForm . new ( %{ company : %{ } , partners : [ %{ } ] } ) iex&gt; form = Spike . make_dirty ( form ) iex&gt; Spike . dirty_fields ( form ) %{ form . ref =&gt; [ :accepts_conditions , :company , :partners ] , form . company . ref =&gt; [ :country , :name ] , hd ( form . partners ) . ref =&gt; [ :name ] }","ref":"Spike.html#make_dirty/1","title":"Spike.make_dirty/1","type":"function"},{"doc":"Given a dirty Spike form, makes it pristine, recursively. iex&gt; form = Test.ComplexForm . new ( %{ company : %{ } , partners : [ %{ } ] } ) iex&gt; form = Spike . make_dirty ( form ) iex&gt; form = Spike . make_pristine ( form ) iex&gt; Spike . dirty_fields ( form ) %{ }","ref":"Spike.html#make_pristine/1","title":"Spike.make_pristine/1","type":"function"},{"doc":"Allows you to update fields in a Spike form that were marked as private. iex&gt; form = Test.PrivateForm . new ( %{ private_field : &quot;foo&quot; } , cast_private : true ) iex&gt; form = Spike . update ( form , form . ref , %{ private_field : &quot;bar&quot; } ) iex&gt; form . private_field &quot;foo&quot; iex&gt; form = Spike . set_private ( form , form . ref , :private_field , &quot;bar&quot; ) iex&gt; form . private_field &quot;bar&quot;","ref":"Spike.html#set_private/4","title":"Spike.set_private/4","type":"function"},{"doc":"Updates a top level Spike form with given params. Marks updated fields as dirty. iex&gt; form = Test.SimpleForm . new ( %{ } ) iex&gt; form = Spike . update ( form , %{ first_name : &quot;Spike&quot; , last_name : &quot;Spiegel&quot; , accepts_conditions : &quot;1&quot; } ) iex&gt; Spike . errors ( form ) %{ }","ref":"Spike.html#update/2","title":"Spike.update/2","type":"function"},{"doc":"Updates a Spike form, by finding a nested form by it's ref field, with given params. Marks updated fields as dirty. iex&gt; form = Test.ComplexForm . new ( %{ company : %{ name : &quot;ACME Corp.&quot; } , partners : [ %{ } ] } ) iex&gt; form . company . name &quot;ACME Corp.&quot; iex&gt; form = Spike . update ( form , form . company . ref , %{ name : &quot;Amazon&quot; } ) iex&gt; form . company . name &quot;Amazon&quot; iex&gt; Spike . dirty_fields ( form ) %{ form . company . ref =&gt; [ :name ] , form . ref =&gt; [ :company ] }","ref":"Spike.html#update/3","title":"Spike.update/3","type":"function"},{"doc":"Returns true if given form has no errors. iex&gt; form = Test.SimpleForm . new ( %{ } ) iex&gt; Spike . valid? ( form ) false iex&gt; form = Spike . update ( form , form . ref , %{ first_name : &quot;Spike&quot; , last_name : &quot;Spiegel&quot; , accepts_conditions : &quot;1&quot; } ) iex&gt; Spike . valid? ( form ) true","ref":"Spike.html#valid?/1","title":"Spike.valid?/1","type":"function"},{"doc":"This function should only be used from within &quot;by&quot; validations, where context of parent forms is required. The first element of returned list will be the top-level form, followed by association name, and another form (if present) followed by assciation name etc. This is useful when in a child form you want to perform a validation that relies on value that lives in a parent form or a sibling in embeds_many .","ref":"Spike.html#validation_context/1","title":"Spike.validation_context/1","type":"function"},{"doc":"Use this module to define your Spike forms. Simple use cases Simple Spike form, with no validation and no nested structs will look like this: defmodule MyApp.RegistrationForm do use Spike.Form do field ( :first_name , :string ) field ( :last_name , :string ) field ( :age , :integer ) field ( :email , :string ) field ( :accepts_conditions , :boolean ) end end form = MyApp.RegistrationForm . new ( %{ first_name : &quot;Spike&quot; } ) form = Spike . update ( form , form . ref , %{ last_name : &quot;Spiegel&quot; } ) form . first_name =&gt; &quot;Spike&quot; form . last_name =&gt; &quot;Spiegel&quot; Adding validations Spike uses Vex as a validation library, so you can add form validations easily: defmodule MyApp.RegistrationForm do use Spike.Form do field ( :first_name , :string ) field ( :last_name , :string ) field ( :age , :integer ) field ( :email , :string ) field ( :accepts_conditions , :boolean ) end validates ( :first_name , presence : true ) validates ( :accepts_conditions , acceptance : true ) end form = MyApp.RegistrationForm . new ( %{ } ) Spike . valid? ( form ) =&gt; false Spike . errors ( form ) [ form . ref ] =&gt; %{ accepts_conditions : [ acceptance : &quot;must be accepted&quot; ] , first_name : [ presence : &quot;must be present&quot; ] } Nested forms with contextual validations You can have nested forms, supproting nested validations as well, where child item can access parent or sibling items by fetching validation context using Spike.validation_context/1 . defmodule MyApp.BudgetPlanner do defmodule LineItem do use Spike.Form do field ( :price , :integer ) field ( :name , :string ) validates ( :name , presence : true ) validates ( :price , presence : true , by : &amp; __MODULE__ . validate_price_within_budget / 2 ) end def validate_price_within_budget ( _price , this_line_item ) do [ parent , :line_items ] = Spike . validation_context ( this_line_item ) sum = parent . line_items |&gt; Enum . reduce_while ( 0 , fn line_item , acc -&gt; if line_item . ref == this_line_item . ref do { :halt , acc + line_item . price } else { :cont , acc + line_item . price } end end ) if parent . max_budget &amp;&amp; sum &gt; parent . max_budget do { :error , &quot;exceeds max budget of \#{ parent . max_budget } &quot; } else :ok end end end use Spike.Form do field ( :max_budget , :integer ) embeds_many ( :line_items , __MODULE__ . LineItem ) end end For functions useful to manipulate forms, see [Spike]. For schema definition look into Spike.Form.Schema . To initialize a Spike form, by casting a map to it's fields (recursively), you can use new/1 callback. form = MyApp.BudgetPlanner . new ( %{ max_budget : 12 , line_items : [ %{ name : &quot;Cheap one&quot; , price : 1 } ] } ) Spike . valid? ( form ) =&gt; true form = Spike . append ( form , form . ref , :line_items , %{ name : &quot;Expensive one&quot; , price : 9000 } ) Spike . valid? ( form ) =&gt; false Spike . human_readable_errors ( form ) =&gt; %{ &quot;line_items.1.price&quot; =&gt; [ &quot;exceeds max budget of 12&quot; ] } In case you need to cast fields marked as private, use new/2 where second parameter is [cast_private: true] .","ref":"Spike.Form.html","title":"Spike.Form","type":"behaviour"},{"doc":"","ref":"Spike.Form.html#c:after_update/3","title":"Spike.Form.after_update/3","type":"callback"},{"doc":"","ref":"Spike.Form.html#c:new/1","title":"Spike.Form.new/1","type":"callback"},{"doc":"","ref":"Spike.Form.html#c:new/2","title":"Spike.Form.new/2","type":"callback"},{"doc":"","ref":"Spike.Form.Schema.html","title":"Spike.Form.Schema","type":"module"},{"doc":"","ref":"Spike.Form.Schema.html#embeds_many/2","title":"Spike.Form.Schema.embeds_many/2","type":"macro"},{"doc":"","ref":"Spike.Form.Schema.html#embeds_one/2","title":"Spike.Form.Schema.embeds_one/2","type":"macro"},{"doc":"Contains macros used to define fields and embeds.","ref":"Spike.Form.Schema.html#field/3","title":"Spike.Form.Schema.field/3","type":"macro"},{"doc":"Spike helps you build stateul, server-memory backed forms in Elixir. If you are struggling with making deep nested Ecto changesets back your forms the way you like it, you may have ended up in a right place.","ref":"readme.html","title":"Readme","type":"extras"},{"doc":"Available in Hex , the package can be installed by adding spike to your list of dependencies in mix.exs : def deps do [ { :spike , &quot;~&gt; 0.2&quot; } ] end Documentation can be found on HexDocs .","ref":"readme.html#installation","title":"Readme - Installation","type":"extras"},{"doc":"Spike can be used on it's own, or with Phoenix LiveView / Surface UI . Basic usage consists on creating an Elixir module, which represents your form and stores both: data, that the user can manipulate using UI controls (via LiveView or otherwise), and data that is necessary for the form to be rendered and displayed to user. Spike's LiveView bindings or Surface UI bindings can be used together with this core library to extend live views or components with out-of-the box support for @form and @errors , as well as default implementation of events handling, to build Spike forms are based on Elixir structs, that declare fields, associations and validations - similar to Ecto schemas or ActiveRecord models. These forms, however, live entirely in the memory. Let's consider a registration form. We have to define a module, with fields and validations: defmodule MyApp.RegistrationForm do use Spike.Form do field ( :first_name , :string ) field ( :last_name , :string ) field ( :age , :integer ) field ( :email , :string ) field ( :accepts_conditions , :boolean , default : false ) end validates ( :first_name , presence : true ) validates ( :accepts_conditions , acceptance : true ) end form = MyApp.RegistrationForm . new ( %{ } ) Spike . valid? ( form ) =&gt; false Spike . errors ( form ) [ form . ref ] =&gt; %{ accepts_conditions : [ acceptance : &quot;must be accepted&quot; ] , first_name : [ presence : &quot;must be present&quot; ] } See the documentation for Spike , Spike.Form , and Spike.Form.Schema modules for usage and examples. For more complete example of usage with LiveView and Surface UI, have a look at our example application .","ref":"readme.html#usage","title":"Readme - Usage","type":"extras"}]